;;;;
;;;; Parsing: Syntactic analysis.
;;;;

;; AST types.

(define-record-type AST (fields location))

(define-record-type Phrase (parent AST) (fields constituents))
(define-record-type Constant (parent AST) (fields value))
(define-record-type Identifier (parent AST) (fields name))
(define-record-type Tuple (parent AST) (fields items))
(define-record-type Array (parent AST) (fields items))
(define-record-type Body (parent AST) (fields phrases))
(define-record-type Block (parent AST) (fields body))
(define-record-type Call (parent AST) (fields head arguments))
(define-record-type Index (parent AST) (fields object indexes))


(define (make-phrase-or-constituent items)
  ;; Build a Phrase node if there is more than one constituent in the sequence;
  ;; otherwise, return the lone constituent iself.
  (cond [(null? items)
         ;; Is this really and EOF error? Clean this up!
         (raise-eof-error 'make-phrase-or-constituent "Phrase with no items")]
        [(null? (cdr items)) (car items)]
        [else (make-Phrase (AST-location (car items)) items)]))

(define (parse-phrase tokens skipnl?)
  ;; Parse a phrase from a list of tokens. When `skipnl?`, ignore newlines.
  ;;
  ;; Return two values: the AST node, and the remaining tokens.
  (let loop ([parts '()] [tokens tokens])
    (let-values ([(token rest) (next-token tokens skipnl?)])
      (cond
       [(and token
             (eqv? (token-value token) #\newline)
             (null? parts))
        ;; Special case: empty line; just skip it.
        (loop parts (cdr tokens))]
       [(end-of-phrase? token skipnl?)
        (values (make-phrase-or-constituent (reverse parts)) tokens)]
       [else
        (let-values ([(part tokens) (parse-constituent tokens skipnl? 0)])
          (loop (cons part parts) tokens))]))))

;; For parsing purposes, a constituent is divided into two parts:
;;
;; - The 'onset', which is the smallest thing that could be a constituent on
;;    its own, such as `2`, or `-x` (because `-` on its own cannot be a
;;    constituent);
;;
;; - If the onset is followed by an operator or an open paren/bracket, then
;;   what follows is the constituent's 'rhyme'. Otherwise, the constituent is
;;   finished.

(define (parse-constituent tokens skipnl? precedence)
  ;; Parse a constituent from a list of tokens. Stop upon finding an operator
  ;; whose precedence is lower than `precedence`, or when no more tokens can
  ;; be added to the constituent.
  ;;
  ;; Return two values: the constituent AST node, and the remaining tokens.
  (let-values ([(left rest) (parse-constituent-onset tokens skipnl?)])
    (parse-constituent-rhyme left rest skipnl? precedence)))

(define (parse-constituent-onset tokens skipnl?)
  ;; Parse a constituent onset (the smallest thing that could be a constituent
  ;; on its own).
  ;;
  ;; Return two values: the AST node, and the remaining tokens.
  (let-values ([(token rest) (next-token tokens skipnl?)])
    (cond
     [(eqv? (token-kind token) 'integer)
      (values (make-Constant (token-location token)
                             (string->number (token-value token)))
              rest)]
     [(eqv? (token-kind token) 'string)
      (values (make-Constant (token-location token)
                             (token-value token))
              rest)]
     [(eqv? (token-kind token) 'symbol)
      (values (make-Constant (token-location token)
                             (string->symbol (token-value token)))
              rest)]
     [(eqv? (token-kind token) 'identifier)
      (values (make-Identifier (token-location token)
                               (string->symbol (token-value token))) rest)]
     [(eqv? (token-kind token) 'operator) ;; unary, prefix
      (let-values ([(arg rest) (parse-constituent (cdr tokens) skipnl?
                                                  (operator-precedence token))])
        (values (make-operator-expr (token-location token)
                                    token
                                    arg)
                rest))]
     [(eqv? (token-value token) #\{)
      (let*-values ([(body rest) (parse-body (token-location token) rest)]
                    [(next rest) (next-token rest #t)])
        (cond [(eqv? (token-value next) #\})
               (values (make-Block (token-location token) body) rest)]
              [(not (token-value next))
               (raise-eof-error 'parse-constituent-onset "Expecting }, found EOF")]
              [else
               (error 'parse-constituent-onset "Expecting }, found" (token-value next))]))]
     [(member (token-value token) '(#\( #\[))
      (let-values ([(items rest) (parse-list token rest)])
        (values (if (eqv? (token-value token) #\()
                    (if (= (length items) 1)
                        ;; parenthesized single element is equal to the element itself
                        (car items)
                      (make-Tuple (token-location token) items))
                  (make-Array (token-location token) items))
                rest))]
     [else (error 'parse-constituent "Found unexpected token" token)])))

(define (parse-constituent-rhyme left tokens skipnl? precedence)
  ;; Parse a constituent rhyme (whatever remains after the onset), and combine
  ;; the onset `left` with it. Stop upon finding an operator whose precedence
  ;; is lower than `precedence`, or when no more tokens can be added to the
  ;; constituent.
  ;;
  ;; Return two values: the resulting AST node, and the remaining tokens.
  (let-values ([(token rest) (next-token tokens skipnl?)])
    (cond
     [(end-of-phrase? token skipnl?)
      (values left tokens)]
     [(and (eqv? (token-kind token) 'operator)
           (> (operator-precedence token) precedence))
      (let-values ([(right rest) (parse-constituent rest skipnl? (operator-precedence token))])
        (parse-constituent-rhyme (make-operator-expr (AST-location left) token left right)
                                 rest skipnl? precedence))]
     [(and (member (token-value token) '(#\( #\[))
           (not (token-spaced? token)))
      (let-values ([(right rest) (parse-list token rest)])
        (parse-constituent-rhyme
         ((if (eqv? (token-value token) #\() make-Call make-Index)
          (token-location token)
          left
          right)
         rest skipnl? precedence))]
     [else (values left tokens)])))

(define (parse-list open tokens)
  ;; Parse a comma-separated list of phrases.
  ;;
  ;; Return two values: the list of AST nodes, and the remaining tokens.
  (define close  (if (eqv? (token-value open) #\() #\) #\]))
  (let loop ([items '()] [tokens tokens])
    (let-values ([(token rest) (next-token tokens #t)])
      (cond
       [(eqv? (token-value token) close)
        (values (reverse items) rest)]
       [else
        (let*-values ([(item rest) (parse-phrase tokens #t)]
                      [(next next-rest) (next-token rest #t)])
          (cond [(eqv? (token-value next) #\,) (loop (cons item items) next-rest)]
                [(eqv? (token-value next) close) (loop (cons item items) rest)]
                [(not (token-value next))
                 (raise-eof-error 'parse-list "Expecting comma or close list, found EOF")]
                [else (error 'parse-list "expecting comma or close list, found" next)]))]))))

(define (make-operator-expr location operator . args)
  ;; Return an AST node for an operator call.
  (make-Call location
             (make-Identifier (token-location operator)
                              (string->symbol (token-value operator)))
             args))

;; Table of operator precedences. Operator precedence is determined by the
;; last character of the operator, according to the table below. The higher
;; the precedence, the tighter the operator binds (e.g., `*` has a higher
;; precedence value than `+`).
(define operator-char-precedences
  '((#\? . 1)
    (#\| . 2)
    (#\& . 3)
    (#\= . 4) (#\< . 4) (#\> . 4)
    (#\: . 5)
    (#\~ . 6)
    (#\+ . 7) (#\- . 7)
    (#\* . 8) (#\/ . 8) (#\% . 8)
    (#\^ . 9)
    (#\@ . 10) (#\! . 10)
    (#\. . 11)
    (#\$ . 12)))

(define (operator-precedence token)
  ;; Return the precedence of an operator token.
  (let* ([operator (token-value token)]
         [last-char (string-ref operator (- (string-length operator) 1))])
    (cdr (assoc last-char operator-char-precedences))))

(define (parse-body location tokens)
  ;; Parse the body of a program or block.
  ;;
  ;; Return two values: the body AST node, and the remaining tokens.
  (let loop ([phrases '()] [tokens tokens])
    (let-values ([(token _) (next-token tokens #t)])
      (cond
       [(member (token-value token) '(#f #\}))
        (values (make-Body location (reverse phrases))
                (if token tokens '()))] ;; discard trailing newlines
       [else
        (let*-values ([(phrase rest) (parse-phrase tokens #f)]
                      [(next next-rest) (next-token rest #f)])
          (loop (cons phrase phrases)
                (if (member (token-value next) '(#\; #\newline))
                    next-rest
                  rest)))]))))
