;;;;
;;;; Evaluation.
;;;;

;; Fenius value types.

;; Some Fenius types have dedicated Scheme types for them.
(define-record-type Nil)
(define nil (make-Nil))
(define-record-type SpecialOperator (fields evaluator))
(define-record-type Macro (fields transformer))

;; Other types have conventional mappings to Scheme types, to ease interoperability:
;;
;; - Fenius numbers, strings, and booleans are directly represented by the
;;   corresponding Scheme types.
;;
;; - Fenius arrays map to Scheme lists. In the future, we may have other, more
;;   efficient representations for lists too.
;;
;; - Fenius tuples map to Scheme vectors.
;;
;; - Scheme procedures can be directly called from Fenius (although a more
;;   elaborate Fenius function type exists too, to handle named arguments
;;   and other aspects of 'native' Fenius functions.
;;
;;   - Multiple values returned by a Scheme procedure are presented as tuples
;;     to the Fenius program. (This means a Fenius program cannot distinguish a
;;     Scheme procedure returning a vector from one returning multiple
;;     values.)

(define-record-type binding (fields name (mutable value)))
(define-record-type env (fields bindings parent))

(define-record-type Module (fields
                            name ;; string!
                            local-bindings
                            imported-bindings))

(define module-map (make-hashtable string-hash string=?))

(define (register-module! module)
  (hashtable-set! module-map (Module-name module) module))

(define (build-env bindings parent)
  (make-env (bindings->hashtable bindings) parent))

(define (bindings->hashtable bindings)
  (let* ([table (make-eq-hashtable)])
    (for-each (lambda (binding)
                (hashtable-set! table (binding-name binding) binding))
              bindings)
    table))

(define (build-module name bindings)
  (make-Module name
               (bindings->hashtable bindings)
               (make-eq-hashtable)))

(define (module-import-contents! importing-module imported-module)
  (hash-table-for-each
   (Module-local-bindings imported-module)
   (lambda (name imported-binding)
     (hashtable-set! (Module-imported-bindings importing-module)
                     name
                     imported-binding))))

(define (env-set! env name value)
  (binding-value-set! (lookup name env) value)
  value)

(define (lookup name namespace)
  (cond [(not namespace)
         (error 'lookup "Unbound variable" name)]
        [(env? namespace)
         (or (hashtable-ref (env-bindings namespace) name #f)
             (lookup name (env-parent namespace)))]
        [(Module? namespace)
         (module-ref namespace name)]
        [else (error 'lookup "Unknown namespace object" namespace)]))

(define (module-ref module name)
  (or (hashtable-ref (Module-local-bindings module) name #f)
      (hashtable-ref (Module-imported-bindings module) name #f)
      (error 'lookup "Unbound variable" name)))

(define (env-add! namespace name value)
  (let ([binding (make-binding name value)])
    (cond [(env? namespace)
           (hashtable-set! (env-bindings namespace) name binding)]
          [(Module? namespace)
           (hashtable-set! (Module-local-bindings namespace) name binding)]
          [else (error 'lookup "Unknown namespace object" namespace)])))

(define base-module (make-parameter (build-module "std/base" '())))

;;;;;;;;;;;;;;;;;

;; Keep track of call stack for debugging.
(define-record-type EvalFrame (fields function ast))
(define call-stack (make-parameter '()))
(define current-function (make-parameter #f))

(define (with-call-stack ast thunk)
  (parameterize ([call-stack (cons (make-EvalFrame (current-function) ast) (call-stack))])
    (thunk)))

(define (evaluate expr env)
  ;; (display "Evaluating: ") (write (record->list expr)) (newline)
  (cond [(Constant? expr)     (Constant-value expr)]
        [(Identifier? expr)   (with-call-stack expr
                                (lambda ()
                                  (binding-value (lookup (Identifier-name expr) env))))]
        [(Block? expr)        (evaluate (Block-body expr) (build-env '() env))]
        [(Body? expr)         (evaluate-phrase-list (Body-phrases expr) env)]
        [(Phrase? expr)       (with-call-stack expr
                                (lambda ()
                                  (evaluate-phrase expr env)))]
        [(Call? expr)         (with-call-stack expr
                                (lambda ()
                                  (evaluate-call expr env)))]
        [(Array? expr)        (evaluate-items (Array-items expr) env)]
        [(Tuple? expr)        (list->vector (evaluate-items (Tuple-items expr) env))]
        [(Index? expr)        (with-call-stack expr
                                (lambda ()
                                  (evaluate-index expr env)))]
        [else                 (error 'evaluate "Unknown object" expr)]))

(define (evaluate-phrase-list phrases env)
  (fold-left (lambda (_ phrase) (evaluate phrase env)) nil phrases))

(define (evaluate-phrase expr env)
  (let* ([parts (Phrase-constituents expr)]
         [head (evaluate (car parts) env)])
    (cond [(SpecialOperator? head)
           ((SpecialOperator-evaluator head) expr env)]
          [(Macro? head)
           (let ([new-expr (fen-apply (Macro-transformer head)
                                      (digest-args (list expr) (lambda (arg) arg)))])
             (evaluate new-expr env))]
          [(null? (cdr parts)) head]
          [else (error 'evaluate "Phrase with multiple constituents, but head is neither a special operator nor a macro\nCommon causes: forgetting a comma between function arguments; leaving a space between function and '('" head)])))

(define (evaluate-call expr env)
  (let ([head (evaluate (Call-head expr) env)])
    (cond [(SpecialOperator? head)
           ((SpecialOperator-evaluator head) expr env)]
          [(Macro? head)
           (let ([new-expr (fen-apply (Macro-transformer head)
                                      (digest-args (list expr) (lambda (arg) arg)))])
             (evaluate new-expr env))]
          [else
           (fen-apply head (digest-args (Call-arguments expr)
                                        (lambda (arg) (evaluate arg env))))])))

(define-record-type Args (fields positional-args keyword-args))

(define (digest-args args evaluator)
  (let loop ([args args] [positional '()] [keyword '()])
    (match args
      ['()
       (make-Args (reverse positional) (reverse keyword))]
      [(list arg & rest)
       (match arg
         [(Call _ (Identifier _ '=) (list (Identifier _ key) value))
          (loop rest positional (cons (cons key (evaluator value)) keyword))]
         [(Call _ (Identifier _ '*) (list items))
          (loop rest (append (reverse (evaluator items)) positional) keyword)]
         [arg
          (loop rest (cons (evaluator arg) positional) keyword)])])))

(define-record-type BuiltinFunction (fields procedure))

(define (fen-apply head args)
  (cond
   [(Function? head)         (fen-apply-function head args)]
   [(BuiltinFunction? head)  ((BuiltinFunction-procedure head) args)]
   [(Class? head)            (fen-apply-class head args)]
   [(procedure? head)        (fen-apply-procedure head args)]
   [else (error 'fen-apply "Not an applicable object" head)]))

(define (fen-apply-procedure proc args)
  (when (not (null? (Args-keyword-args args)))
    (error 'fen-apply "Native Scheme procedure cannot be called with keyword arguments" proc))
  (call-with-values
      (lambda ()
        (apply proc (Args-positional-args args)))
    (case-lambda
      [()       nil]
      [(value)  value]
      [values   (list->vector values)])))

(define (fen-apply-gen-bindings fun args)
  (match fun
    [(Function name (FnSignature params rest-args rest-kwargs) _body env)
     (let loop ([params params]
                [positional (Args-positional-args args)]
                [keyword (Args-keyword-args args)]
                [bindings '()])
       (match params
         ['() (append
               (cond [rest-args (list (make-binding rest-args positional))]
                     [(null? positional) '()]
                     [else (error 'fen-build-apply-env "Extra positional arguments")])
               (cond [rest-kwargs (list (make-binding rest-kwargs keyword))]
                     [(null? keyword) '()]
                     [else (error 'fen-build-apply-env "Extra keyword arguments")])
               bindings)]
         [(list (ParamDescriptor name optional? default) & other-params)
          (cond [(not (null? positional))
                 (loop other-params (cdr positional) keyword
                       (cons (make-binding name (car positional)) bindings))]
                [(assoc name keyword)
                 => (lambda (key+val)
                      (loop other-params positional (cdr keyword)
                            (cons (make-binding name (cdr key+val)) bindings)))]
                [optional?
                 (loop other-params positional keyword
                       (cons (make-binding name (evaluate default env)) bindings))]
                [else (error 'fen-build-apply-env "Param not provided" name)])]))]))

(define (fen-apply-function fun args)
  (let ([env (build-env (fen-apply-gen-bindings fun args)
                        (Function-env fun))])
    (evaluate (Function-body fun) env)))



(define (fen-apply-class head args)
  (when (not (null? (Args-keyword-args args)))
    (error 'fen-apply-class "Class constructors currently don't support keyword arguments"))
  (let* ([rtd (Class-rtd head)]
         [constructor (record-constructor
                       (make-record-constructor-descriptor rtd #f #f))])
    (apply constructor (Args-positional-args args))))



(define (evaluate-items items env)
  (map (lambda (item) (evaluate item env)) items))

(define (evaluate-index expr env)
  ;; Kludge: translate to method call to 'at'. The proper thing would
  ;; be to make the call without explicitly building a synthetic
  ;; syntax tree to evaluate.
  (evaluate
   (make-Call (AST-location expr)
              (make-Identifier (AST-location expr) '\.)
              (list (Index-object expr)
                    (make-Call (AST-location expr)
                               (make-Identifier (AST-location expr) 'at)
                               (Index-indexes expr))))
   env))

;;;;
;;;; Special operators
;;;;

(define specialop-let
  (make-SpecialOperator
   (lambda (expr env)
     (match expr
       [(Phrase _ (list _let (Call _ (Identifier _ '=) (list lhs rhs))))
        (match lhs
          [(Identifier _ name)
           (let ([result (evaluate rhs env)])
             (env-add! env name result)
             result)]
          [(Call _ (Identifier _ name) params)
           (let ([fun (make-function name params rhs env)])
             (env-add! env name fun)
             fun)]
          [else (error 'let "Unknown left-hand side" lhs)])]
       [other (error 'let "Unknown let format" other)]))))

;;;; Functions.

(define-record-type ParamDescriptor (fields name optional? default))
(define-record-type FnSignature (fields names all-args all-kwargs))

(define-record-type Function (fields name signature body env))

(define (rest-param? param operator)
  (and (Call? param)
       (eqv? (Identifier-name (Call-head param)) operator)
       (Identifier-name (car (Call-arguments param)))))

(define (parse-fn-parameters params)
  (let* ([paramlist (cond [(Tuple? params) (Tuple-items params)]
                          [(list? params) params]
                          [else (list params)])])
    (let loop ([params paramlist] [descriptors '()] [all-args #f] [all-kwargs #f])
      (if (null? params)
          (make-FnSignature (reverse descriptors) all-args all-kwargs)
        (let ([param (car params)])
          (cond [(rest-param? param '*)
                 => (lambda (name)
                      (if all-args
                          (error 'parse-fn-parameters "Multiple *args in signature")
                        (loop (cdr params) descriptors name all-kwargs)))]
                [(rest-param? param '**)
                 => (lambda( name)
                      (if all-kwargs
                          (error 'parse-fn-parameters "Multiplw **kwargs in signature")
                        (loop (cdr params) descriptors all-args name)))]
                [else
                 (let ([descriptor
                        (cond [(Identifier? param)
                               (make-ParamDescriptor (Identifier-name param) #f #f)]
                              [(and (Call? param)
                                    (Identifier? (Call-head param))
                                    (eqv? (Identifier-name (Call-head param)) '=))
                               (let ([name (Identifier-name (car (Call-arguments param)))]
                                     [default (cadr (Call-arguments param))])
                                 (make-ParamDescriptor name #t default))]
                              [else
                               (error 'parse-fn-parameters "Invalid parameter format" param)])])
                   (loop (cdr params) (cons descriptor descriptors) all-args all-kwargs))]))))))

(define (parse-call-arguments arglist)
  (let loop ([args arglist] [positional '()] [named '()])
    (if (null? args)
        (values (reverse positional) (reverse named))
      (let ([arg (car args)])
        (cond [(and (Call? arg)
                    (Identifier? (Call-head arg))
                    (eqv? (Identifier-name (Call-head arg)) '=))
               (let ([name (Identifier-name (car (Call-arguments arg)))]
                     [value (cadr (Call-arguments arg))])
                 (loop (cdr args) positional (cons (cons name value) named)))]
              [else
               (loop (cdr args) (cons arg positional) named)])))))

(define (make-call-env fn-env fn-sig arg-evaluator positional named)
  (let loop ([descriptors (FnSignature-names fn-sig)]
             [bindings '()]
             [positional positional]
             [named named])
    (if (null? descriptors)
        (let ([full-bindings
               (append
                (cond [(FnSignature-all-args fn-sig)
                       (list (make-binding (FnSignature-all-args fn-sig) positional))]
                      [(null? positional) '()]
                      [else (error 'make-call-env "Unconsumed positional arguments" positional)])
                (cond [(FnSignature-all-kwargs fn-sig)
                       (list (make-binding (FnSignature-all-kwargs fn-sig)
                                           (map (lambda (pair) (vector (car pair) (cdr pair))) named)))]
                      [(null? named) '()]
                      [else (error 'make-call-env "Unconsumed named arguments" named)])
                bindings)])
          (build-env full-bindings fn-env))
      (let ([descriptor (car descriptors)])
        (cond [(not (null? positional))
               (loop (cdr descriptors)
                     (cons (make-binding (ParamDescriptor-name descriptor)
                                         (arg-evaluator (car positional)))
                           bindings)
                     (cdr positional)
                     named)]
              [(assoc (ParamDescriptor-name descriptor) named)
               => (lambda (result)
                    (loop (cdr descriptors)
                          (cons (make-binding (ParamDescriptor-name descriptor)
                                              (arg-evaluator (cdr result)))
                                bindings)
                          positional
                          (remq result named)
                          ))]
              [(ParamDescriptor-optional? descriptor)
               (loop (cdr descriptors)
                     (cons (make-binding (ParamDescriptor-name descriptor)
                                         (evaluate (ParamDescriptor-default descriptor) fn-env))
                           bindings)
                     positional
                     named)]
              [else
               (error 'make-call-env "Value not provided for parameter"
                      (ParamDescriptor-name descriptor))])))))


(define specialop-fn
  (make-SpecialOperator
   (lambda (expr env)
     (let* ([params (cadr (Phrase-constituents expr))]
            [body (caddr (Phrase-constituents expr))])
       (when (not (null? (cdddr (Phrase-constituents expr))))
         (error 'fn "Usage: fn (PARAM, ...) EXPR"))
       (make-function #f params body env)))))

(define (make-function name params body env)
  (let ([paramlist (cond [(Tuple? params) (Tuple-items params)]
                         [(list? params) params]
                         [else (list params)])])
    (make-Function name
                   (parse-fn-parameters paramlist)
                   body
                   env)))
    ;; (lambda args
    ;;   (when (not (= (length args) (length paramlist)))
    ;;     (error 'fn-apply "Expected ~s arguments, got ~s" (length paramlist) (length args)))
    ;;   (let ([newenv (build-env (map make-binding paramlist args) env)])
    ;;     (evaluate body newenv)))))

;;;; If.

(define specialop-if
  (make-SpecialOperator
   (lambda (expr env)
     (let loop ([clauses (parse-if-clauses (Phrase-constituents expr))])
       (if (null? clauses)
           nil
         (let ([clause (car clauses)])
           (if (or (eqv? (car clause) #t)
                   (evaluate (car clause) env))
               (evaluate (cadr clause) env)
             (loop (cdr clauses)))))))))

(define (parse-if-clauses items)
  (let loop ([clauses '()] [items items])
    (cond [(null? items)
           (if (null? clauses)
               (error 'parse-if-clauses "Incomplete if command")
             (reverse clauses))]
          [(Identifier? (car items))
           (cond [(or (and (null? clauses)
                           (eqv? (Identifier-name (car items)) 'if))
                      (and (not (null? clauses))
                           (eqv? (Identifier-name (car items)) 'elif)))
                  (let ([test (cadr items)]
                        [next (caddr items)])
                    (if (and (Identifier? next)
                             (eqv? (Identifier-name next) 'then))
                        (loop (cons (list test (cadddr items)) clauses)
                              (cddddr items))
                      (loop (cons (list test (caddr items)) clauses)
                            (cdddr items))))]
                 [(eqv? (Identifier-name (car items)) 'else)
                  (loop (cons (list #t (cadr items)) clauses)
                        (cddr items))]
                 [else (error 'parse-if-clauses "Unexpected word" (car items))])]
          [else
           (error 'parse-if-clauses "Expecting if/elif/else" (car items))])))

;;;; loops

(define specialop-while
  (make-SpecialOperator
   (lambda (expr env)
     (let ([test (cadr (Phrase-constituents expr))]
           [body (caddr (Phrase-constituents expr))])
       (assert (null? (cdddr (Phrase-constituents expr))))
       (let loop ()
         (cond [(evaluate test env)
                (evaluate body env)
                (loop)]
               [else nil]))))))

;;;; 'do' blocks.

;; do BLOCKNAME (var=value, ...) BODY

(define-record-type DoBlockDescriptor (fields tag name fn))
(define-record-type RedoException (fields tag args))
(define-record-type ReturnException (fields tag value))

(define specialop-do
  (make-SpecialOperator
   (lambda (expr env)
     (match expr
       [(Phrase _ (list _do (Call _ name params) body))
        (let* ([fn (make-function (Identifier-name name)
                                  (cons name params)
                                  body
                                  env)]
               [tag (make-DoBlockDescriptor (gensym) name fn)])
          (let loop ([args (digest-args '() (lambda (arg) (evaluate arg env)))])
            (guard (exc [(and (RedoException? exc)
                              (eqv? tag (RedoException-tag exc)))
                         (loop (RedoException-args exc))]
                        [(and (ReturnException? exc)
                              (eqv? tag (ReturnException-tag exc)))
                         (ReturnException-value exc)])
              (fen-apply-function fn (make-Args (cons tag (Args-positional-args args))
                                                (Args-keyword-args args))))))]))))

(define specialop-redo
  (make-SpecialOperator
   (lambda (expr env)
     (match expr
       [(Phrase _ (list _redo (Call _ (Identifier _ name) args)))
        (let* ([tag (binding-value (lookup name env))])
          (raise (make-RedoException tag (digest-args args (lambda (arg) (evaluate arg env))))))]))))

(define specialop-return
  (make-SpecialOperator
   (lambda (expr env)
     (match expr
       [(Phrase _ (list _return value (Identifier _ 'from) (Identifier _ name)))
        (let* ([tag (binding-value (lookup name env))])
          (assert (DoBlockDescriptor? tag))
          (raise (make-ReturnException tag (evaluate value env))))]))))

;;;;

(define specialop-quote
  (make-SpecialOperator
   (lambda (expr env)
     (car (Call-arguments expr)))))


;;;; Accessors

(define specialop-dot
  (make-SpecialOperator
   (lambda (expr env)
     (let* ([left (car (Call-arguments expr))]
            [right (cadr (Call-arguments expr))])
       (cond [(Identifier? right)
              (let* ([object (evaluate left env)]
                     [accessor (Identifier-name right)]
                     [class (fenius-class-of object)])
                (cond [(Module? object)
                       (binding-value (module-ref object accessor))]    ;; hmmm...
                      [class ((get-class-method class accessor) object)]
                      [else (error 'specialop-dot "Object has no attributes" object)]))]
             ;; kludge: correct for relative precedence of dot and call/index operations.
             [(or (Call? right) (Index? right))
              (let-values ([(get-accessor get-arguments make-expr)
                            (if (Call? right)
                                (values Call-head Call-arguments make-Call)
                              (values Index-object Index-indexes make-Index))])
                (evaluate (make-expr (AST-location expr)
                                     (make-Call (AST-location (Call-head expr))
                                                (Call-head expr) ;; '.'
                                                (list left (get-accessor right)))
                                     (get-arguments right))
                          env))]
             [else (error 'specialop-dot "Unknown rhs" right)])))))


;; Record definition

(define (parse-record-declaration expr)
  (match expr
    [(Phrase _ (list _record (Call _ (Identifier _ '>)
                                     (list parent
                                           (Call _ (Identifier _ name) args)))))
     (values name parent args)]
    [(Phrase _ (list _record (Call _ (Identifier _ name) args)))
     (values name #f args)]
    ))


(define specialop-record
  (make-SpecialOperator
   (lambda (expr env)
     (let-values ([(name parent args) (parse-record-declaration expr)])
       (let* ([fields (map Identifier-name args)]
              [parent-rtd (and parent
                               (Class-rtd (evaluate parent env)))]
              [rtd (make-record-type-descriptor
                    name parent-rtd #f #f #f
                    (list->vector (map (lambda (field) (list 'mutable field)) fields)))]
              [class (fenius-rtd-class rtd)])
         (env-add! env name class)
         class)))))

(define (record-type-all-accessors rtd)
  (cond [(not rtd) '()]
        [else (append (record-type-all-accessors (record-type-parent rtd))
                      (let* ([fields (record-type-field-names rtd)]
                             [numfields (vector-length fields)])
                        (let loop ([i 0])
                          (if (>= i numfields)
                              '()
                            (cons (list (vector-ref fields i) (record-accessor rtd i))
                                  (loop (+ i 1)))))))]))

;;;; Types / classes.

(define record-class-map (make-eq-hashtable))

(define-record-type Class
  (fields name
          parent
          accessors
          mutators
          methods
          rtd))

(define (make-empty-class name parent)
  (make-Class name
              parent
              (make-eq-hashtable)
              (make-eq-hashtable)
              (make-eq-hashtable)
              #f))

(define (add-method! class methodname proc)
  (hashtable-set! (Class-methods class) methodname proc))

(define class-Object (make-empty-class 'Object #f))
(define class-Number (make-empty-class 'Number class-Object))
(define class-Bool (make-empty-class 'Bool class-Object))
(define class-Symbol (make-empty-class 'Symbol class-Object))
(define class-List (make-empty-class 'List class-Object))
(define class-Tuple (make-empty-class 'Tuple class-Object))
(define class-String (make-empty-class 'String class-Object))
(define class-Unknown (make-empty-class 'Unknown class-Object))

(define class-Record (make-empty-class 'Record class-Object))

(hashtable-set! (Class-accessors class-List) 'first car)
(hashtable-set! (Class-accessors class-List) 'rest cdr)
(hashtable-set! (Class-accessors class-List) 'length length)

(define-record-type Range (fields start end inclusive?))

(add-method! class-List 'cons
             (lambda (lst)
               (lambda (item)
                 (cons item lst))))

;; at and at_setter methods for various objects.

(add-method! class-List 'at
             (lambda (obj)
               (lambda (index)
                 (list-ref obj index))))

(add-method! class-List 'at_setter
             (lambda (obj)
               (lambda (value index)
                 (set-car! (list-tail obj index) value)
                 value)))


(add-method! class-Tuple 'at
             (lambda (obj)
               (lambda (index)
                 (vector-ref obj index))))

(add-method! class-Tuple 'at_setter
             (lambda (obj)
               (lambda (value index)
                 (vector-set! obj index value)
                 value)))

;; TODO: String indexing is by Unicode codepoint? by byte? by
;; grapheme? Does it return a string? a Character type (which we
;; currently don't have)? For now, we just mimic the Scheme host.
(add-method! class-String 'at
             (lambda (obj)
               (lambda (index)
                 (cond [(integer? index) (string-ref obj index)]
                       [(Range? index)
                        (substring obj
                                   (Range-start index)
                                   (+ (Range-end index)
                                      (if (Range-inclusive? index) 1 0)))]
                       [else (error 'string/at "Not a valid index" index)])
                 )))

;; Strings should not even be mutable.
(add-method! class-String 'at_setter
             (lambda (obj)
               (lambda (value index)
                 (string-set! obj index value)
                 value)))


;; Crazier thing: everything has a 'cons' method, so 'cons' appears infix.
;; Seemed like a good idea at the time...
;; (hashtable-set! (Class-accessors class-Object) 'cons
;;                 (lambda (item)
;;                   (lambda (lst)
;;                     (cons item lst))))

(define (get-class-method class name)
  (if (not class)
      (error 'get-class-method "Unknown accessor or method" name)
    (or (hashtable-ref (Class-methods class) name #f)
        (hashtable-ref (Class-accessors class) name #f)
        (get-class-method (Class-parent class) name))))

(define (get-class-mutator class name)
  (if (not class)
      (error 'get-class-mutator "Unknown mutator" name)
    (or (hashtable-ref (Class-mutators class) name #f)
        (get-class-mutator (Class-parent class) name))))

(define (fenius-class-of val)
  (cond [(number? val) class-Number]
        [(boolean? val) class-Bool]
        [(symbol? val) class-Symbol]
        [(list? val) class-List]
        [(vector? val) class-Tuple]
        [(string? val) class-String]
        [(record? val) (fenius-rtd-class (record-rtd val))]
        [else class-Unknown]))

(define (fenius-rtd-class rtd)
  (or (hashtable-ref record-class-map rtd #f)
      (generate-record-class rtd)))

(define (generate-record-class rtd)
  (let* ([class (make-Class (record-type-name rtd)
                            (cond [(record-type-parent rtd) => fenius-rtd-class]
                                  [else class-Record])
                            (make-eq-hashtable)
                            (make-eq-hashtable)
                            (make-eq-hashtable)
                            rtd)]
         [fields (record-type-field-names rtd)]
         [numfields (vector-length fields)])
    (let loop ([i 0])
      (when (< i numfields)
        (hashtable-set! (Class-accessors class) (vector-ref fields i) (record-accessor rtd i))
        (when (record-field-mutable? rtd i)
          (hashtable-set! (Class-mutators class) (vector-ref fields i) (record-mutator rtd i)))
        (loop (+ i 1))))
    (hashtable-set! record-class-map rtd class)
    class))

(define (fenius-instance-of? obj class)
  (fenius-subclass-of? (fenius-class-of obj) class))

(define (fenius-subclass-of? subclass superclass)
  (cond [(eqv? subclass superclass) #t]
        [(not subclass) #f]
        [else (fenius-subclass-of? (Class-parent subclass) superclass)]))


(define specialop-method
  (make-SpecialOperator
   (lambda (expr env)
     (match expr
       [(Phrase _ (list _method
                        (Call _ (Identifier _ '=)
                              (list
                               (Call _ (Identifier _ '\.)
                                     (list class-name
                                           (Call _ (Identifier _ method-name) method-params)))
                               body))))
        (let ([class (evaluate class-name env)])
          (hashtable-set! (Class-methods class) method-name
                          (lambda (obj)
                            ;; TODO: find a way to create the function
                            ;; only once, and pass 'this' some other way.
                            (make-Function method-name
                                           (parse-fn-parameters method-params)
                                           body
                                           (build-env (list
                                                       (make-binding 'this obj))
                                                      env)))))]))))

;;;; Mutation.

(define specialop-assignment
  (make-SpecialOperator
   (lambda (expr env)
     (let* ([lhs (car (Call-arguments expr))]
            [rhs (cadr (Call-arguments expr))])
       (cond [(Identifier? lhs)
              (env-set! env (Identifier-name lhs) (evaluate rhs env))]
             [(and (Call? lhs)
                   (Identifier? (Call-head lhs))
                   (eqv? (Identifier-name (Call-head lhs)) '\.))
              (let* ([object (evaluate (car (Call-arguments lhs)) env)]
                     [field (Identifier-name (cadr (Call-arguments lhs)))]
                     [mutator (get-class-mutator (fenius-class-of object) field)])
                (if mutator
                    (mutator object (evaluate rhs env))
                  (error ':= "Field has no mutator" object field)))]
             [(Index? lhs)
              ;; Kludge: translate to method call to 'at_setter' (see evaluate-index).
              (evaluate (make-Call (AST-location lhs)
                                   (make-Identifier (AST-location lhs) '\.)
                                   (list (Index-object lhs)
                                         (make-Call (AST-location lhs)
                                                    (make-Identifier (AST-location lhs) 'at_setter)
                                                    (cons rhs (Index-indexes lhs)))))
                        env)]
             [else (error ':= "Unknown lhs" lhs)])))))

;;;; Exceptions.

(define specialop-try
  (make-SpecialOperator
   (lambda (expr env)
     (match expr
       [(Phrase _ (list (Identifier _ 'try) body & rest))
        (let-values ([(handlers finalizer) (parse-except-finally rest env)])
          (dynamic-wind (lambda () #t)
                        (lambda ()
                          (guard (exc [#t (evaluate-try-handlers exc handlers env)])
                            (evaluate body env)))
                        (lambda ()
                          (when finalizer
                            (evaluate finalizer env)))))]))))

(define-record-type HandlerClause (fields var class body))

(define (parse-except-finally exprs env)
  (let loop ([exprs exprs] [handlers '()] [finalizer #f])
    (match exprs
      [(list (Identifier _ 'except) pattern body & rest)
       (let-values ([(var class)
                     (match pattern
                       [(Identifier _ var) (values var class-Object)]
                       [(Call _ (Identifier _ ':) (list (Identifier _ var)
                                                        class))
                        (values var (evaluate class env))]
                       [else (error 'parse-catch-finally "Invalid except pattern" pattern)])])
         (loop rest (cons (make-HandlerClause var class body) handlers) finalizer))]
      [(list (Identifier _ 'finally) body & rest)
       (cond [finalizer
              (error 'parse-except-finally "Multiple 'finally' clauses")]
             [(not (null? rest))
              (error 'parse-except-finally "Trailing clauses after 'finally' clause")]
             [else (loop rest handlers body)])]
      [(list) (values (reverse handlers) finalizer)])))

(define (evaluate-try-handlers exception handlers env)
  (match handlers
    ['() (raise exception)]
    [(list (HandlerClause var class body) & rest)
     (if (fenius-instance-of? exception class)
         (evaluate body (build-env (list (make-binding var exception)) env))
       (evaluate-try-handlers exception rest env))]))

;; XXX: Should this really be a special operator?
;; Technically, it does not have to be, but aestetically feels better being one.
(define specialop-raise
  (make-SpecialOperator
   (lambda (form env)
     (match form
       [(Phrase _ (list (Identifier _ 'raise) val))
        (raise (evaluate val env))]))))

;;;; Module importing.

(define specialop-import
  (make-SpecialOperator
   (lambda (expr env)
     (match expr
       [(Phrase _ (list _import modpath))
        (do-import env modpath #f)]
       [(Phrase _ (list _import modpath (Identifier _ 'as) (Identifier _ importname)))
        (do-import env modpath importname)]))))

(define (do-import env modpath importname)
  (let* ([pathlist (modpath->list modpath)]
         [name (or importname (string->symbol (car (last-pair pathlist))))]
         [module (find-module pathlist)])
    (env-add! env name module)))

(define (modpath->list modpath)
  (match modpath
    [(Constant _ value)
     (cond [(string? value) (string-split value #\/)]
           [(symbol? value) (list (symbol->string value))]
           [else (error 'modpath->list "Invalid module component" value)])]
    [(Identifier _ name)
     (list (symbol->string name))]
    [(Call _ (Identifier _ '/) (list left right))
     (append (modpath->list left) (modpath->list right))]
    [other (error 'modpath-<list "Invalid module component" other)]))

(define (find-module pathlist)
  (let ([name (string-join pathlist "/")])
    (or (hashtable-ref module-map name #f)
        (let ([module (or (find-fenius-module name)
                          (find-scheme-module pathlist))])
          (when (not module)
            (error 'find-module "Module not found" name))
          (hashtable-set! module-map name module) ;; FIXME: should set this before running the module body
          module))))

(define (find-fenius-module name)
  (let ([filename (string-append name ".fen")]) ;; TODO: module search path
    (if (file-exists? filename)
        (load-fenius-module-from-file name filename)
      #f)))

(define (load-fenius-module-from-file modname filename)
  (call-with-input-file filename
    (lambda (infile)
      (let ([source (build-source filename (get-string-all infile))]
            [module (build-module modname '())])
        (module-import-contents! module (base-module))
        (run-fenius-source source module)
        module))))

(define (run-fenius-source source env)
  (let-values ([(body rest) (parse-body (make-location source 0)
                                        (parse-source-tokens source))])
    (let ([result (run-fenius-body body env)])
      (when (not (null? rest))
        (display "Extra tokens: ")
        (write rest)
        (newline))
      result)))

(define (run-fenius-body body env)
  (with-exception-handler
      (lambda (cnd)
        (display "Ouch!\n")
        (display-call-stack (call-stack))
        (raise cnd))
    (lambda ()
      (evaluate body env))))

(define (display-call-stack stack)
  (let loop ([items (reverse stack)])
    (when (not (null? items))
      (let ([item (car items)])
        (let ([location (AST-location (EvalFrame-ast item))])
          (display (if location (source-filename (location-source location)) #f))
          (display ":")
          (display (if location (location-position location) #f))
          (display " ")
          (display (if (EvalFrame-function item) (Function-name (EvalFrame-function item)) #f))
          (newline)

          (when location
            (print-line-in-source (location-source location)
                                  (location-position location)))
          ))
      (loop (cdr items)))))

(define (find-scheme-module pathlist)
  (let* ([modname (map string->symbol pathlist)]
         [env (guard (err [(assertion-violation? err) #f])
                (environment modname))])
    (if (not env)
        #f
      (let ([module (build-module (string-join pathlist "/") '())])
        (for-each (lambda (name)
                    (guard (err [(syntax-violation? err) #f]) ;; skip macros
                      (env-add! module (scheme->fenius-name name) (eval name env))))
                  (environment-symbols env))
        module))))

(define (scheme->fenius-name name)
  (let* ([oldname (symbol->string name)]
         [prefix (case (string-ref oldname (- (string-length oldname) 1))
                   [(#\?) '("is_")]
                   [(#\!) '("do_")]
                   [else '()])])
    (let loop ([oldname (string->list
                         (if (null? prefix)
                             oldname
                           (substring oldname 0 (- (string-length oldname) 1))))]
               [newname prefix])
      (match oldname
        [(list '#\- '#\> & rest)  (loop rest (cons "_to_" newname))]
        [(list '#\- & rest)       (loop rest (cons "_" newname))]
        [(list char & rest)       (loop rest (cons (string char) newname))]
        ['()  (string->symbol (apply string-append (reverse newname)))]))))

;;;;
;;;; Printing.
;;;;

(define write-field-names? (make-parameter #f))

(add-method! class-Object 'repr_to_port
  (lambda (this)
    (lambda (port)
      (write this port))))

(add-method! class-Bool 'repr_to_port
  (lambda (this)
    (lambda (port)
      (display (if this "true" "false") port))))

(add-method! (fenius-rtd-class (record-type-descriptor Nil)) 'repr_to_port
  (lambda (this)
    (lambda (port)
      (display "nil" port))))

(add-method! (fenius-rtd-class (record-type-descriptor Function)) 'repr_to_port
  (lambda (this)
    (lambda (port)
      (fprintf port "<function `~a`>" (Function-name this)))))

(add-method! (fenius-rtd-class (record-type-descriptor hashmap)) 'repr_to_port
  (lambda (this)
    (lambda (port)
      (fenius-write-hashmap this port))))

(add-method! (fenius-rtd-class (record-type-descriptor Class)) 'repr_to_port
  (lambda (this)
    (lambda (port)
      (fprintf port "<class `~a`>" (Class-name this)))))

(add-method! class-Record 'repr_to_port
  (lambda (this)
    (lambda (port)
      (fenius-write-record this port))))

(add-method! class-List 'repr_to_port
  (lambda (this)
    (lambda (port)
      (fenius-write-sequence #\[ #\] this port))))

(add-method! class-Tuple 'repr_to_port
  (lambda (this)
    (lambda (port)
      (fenius-write-sequence #\( #\) (vector->list this) port))))

(add-method! class-Symbol 'repr_to_port
  (lambda (this)
    (lambda (port)
      (fprintf port "`~a`" this))))

(define (fenius-write obj port)
  (fen-apply ((get-class-method (fenius-class-of obj) 'repr_to_port) obj)
             (digest-args (list port) (lambda (arg) arg))))

(define (fenius-write-record obj port)
  (let* ([rtd (record-rtd obj)]
         [name (record-type-name rtd)]
         [all-fields (record-type-all-accessors rtd)])
    (display name port)
    (display "(" port)
    (let loop ([fields all-fields])
      (match fields
        ['() #f]
        [(list (list name accessor) & rest)
         (when (not (eqv? fields all-fields)) (display ", " port))
         (when (write-field-names?)
          (display name port)
          (display "=" port))
         (fenius-write (accessor obj) port)
         (loop rest)]))
    (display ")" port)))

(define (fenius-write-sequence open close all-items port)
  (display open port)
  (let loop ([items all-items])
    (when (not (null? items))
      (when (not (eq? items all-items))
        (display ", " port))
      (fenius-write (car items) port)
      (loop (cdr items))))
  (display close port))

(define (fenius-display obj)
  (cond [(string? obj) (display obj)]
        [else (fenius-write obj (current-output-port))]))

;;;;;

;; record_value.with(field1=value1, ...)
(add-method! class-Record 'with
  (lambda (this)
    (make-BuiltinFunction
     (lambda (args)
       (when (not (null? (Args-positional-args args)))
         (error 'with "with() takes only keyword arguments"))
       (let* ([rtd (record-rtd this)]
              [fields
               (map (lambda (field)
                     (match field
                       [(list name accessor)
                        (cond [(assoc name (Args-keyword-args args))
                               => cdr]
                              [else (accessor this)])]))
                    (record-type-all-accessors rtd))]
              [constructor (record-constructor
                            (make-record-constructor-descriptor rtd #f #f))])
         (apply constructor fields))))))
