;;;;
;;;; Parsing: Lexical analysis.
;;;;

(define-condition-type &source-eof-error &condition
  make-source-eof-error source-eof-error?)

(define (raise-eof-error who message . irritants)
  (raise (condition (make-who-condition who)
                    (make-message-condition message)
                    (make-irritants-condition irritants)
                    (make-source-eof-error))))

;; Token type.
(define-record-type token
  (fields location  ;; position in source string
          kind      ;; kind of token (integer, identifier, operator, etc.)
          value*    ;; contents of the token
          spaced?   ;; is the token preceded by whitespace? This is relevant to distinguish
                    ;; function calls `f(x,y)` from identifier-followed-by-tuple `f (x,y)`.
          ))
(define (token-value token) (if token (token-value* token) #f))

(define-record-type source
  (fields filename text line-positions))

(define (build-source filename text)
  (make-source filename text (find-line-positions text)))

(define (find-line-positions text)
  (let ([length (string-length text)])
    (let loop ([i 0] [positions '()])
      (if (>= i length)
          (list->vector (reverse (cons length positions)))
        (let ([next-newline (string-index #\newline text i)])
          (loop (+ next-newline 1) (cons i positions)))))))

(define (find-line-in-source source position)
  (let* ([positions (source-line-positions source)]
         [limit (vector-length positions)])
    (let loop ([i 0])
      (cond [(>= i limit)
             (error 'find-line-in-source "Position beyond EOF" position positions)]
            [(<  position (vector-ref positions i))
             (let* ([start (vector-ref positions (- i 1))]
                    [end (vector-ref positions i)]
                    [line (substring (source-text source)
                                     start
                                     (if (eqv? #\newline (string-ref (source-text source) (- end 1)))
                                         (- end 1)
                                       end))])
               (values i (- position start) line))]
            [else (loop (+ i 1))]))))

(define (print-line-in-source source position)
  (let-values ([(line column text) (find-line-in-source source position)])
    (display (string-pad (number->string line) 6))
    (display ": ")
    (display text)
    (newline)
    (display (string-pad "^" (+ 9 column)))
    (newline)))

(define (string-pad string length)
  (let ([pad-size (- length (string-length string))])
    (if (<= pad-size 0)
        string
      (string-append (make-string pad-size #\space) string))))


(define-record-type location
  (fields source position))

(define delimiter-chars (string->list "()[]{},;\n"))
(define (delimiter-char? char)
  (and (member char delimiter-chars) #t))

(define operator-chars (string->list "!@$%^&*-+=:.<>/?\\|~"))
(define (operator-char? char)
  (and (member  char operator-chars) #t))

(define (identifier-char? char)
  (or (char-alphabetic? char)
      (char-numeric? char)
      (eqv? char #\_)))

(define (parse-token source text start spaced?)
  ;; Parse the next token in `text`, starting at position `start`. The
  ;; resulting token will have the specified `spaced?` value.
  ;;
  ;; Returns two values: the resulting token (or 'eof), and the next position
  ;; after the token.
  (if (>= start (string-length text))
      (values 'eof start)
    (let ([char (string-ref text start)])
      (cond
       [(eqv? char #\#)
        (parse-token source text (string-index #\newline text start) #t)]
       [(delimiter-char? char)
        (values (make-token (make-location source start) 'delimiter char spaced?)
                (+ 1 start))]
       [(char-whitespace? char) (parse-token source text (+ 1 start) #t)]
       [(operator-char? char)   (parse-token-while operator-char? 'operator source text start spaced?)]
       [(char-numeric? char)    (parse-token-while char-numeric? 'integer source text start spaced?)]
       [(identifier-char? char) (parse-token-while identifier-char? 'identifier source text start spaced?)]
       [(eqv? char #\")         (parse-string source text start spaced?)]
       [(eqv? char #\`)         (parse-symbol source text start spaced?)]
       [else
        ;;(values 'error start)
        (error 'parse-token "Invalid token at" start)]))))

(define (parse-token-while ok? kind source text start spaced?)
  ;; Build a token with the specified `kind` and `spaced?` properties, and
  ;; whose value is the longest sequence of characters that satisfy the
  ;; predicate `ok?`, beginning at `start`.
  ;;
  ;; Return two values: the token, and the next position.
  (let-values ([(value end) (collect-chars-while ok? text start)])
    (values (make-token (make-location source start) kind value spaced?) end)))

(define (collect-chars-while ok? text start)
  ;; Return a string made of the longest sequence of characters that satisfy
  ;; the predicate `ok?`, beginning at `start`.
  ;;
  ;; Return two values: the string, and the next position.
  (let ([length (string-length text)])
    (let loop ([end start])
      (cond [(or (>= end length)
                 (not (ok? (string-ref text end))))
             (values (substring text start end) end)]
            [else (loop (+ 1 end))]))))

(define (parse-source-tokens source)
    ;; Parse all tokens in `text`, beginning at `start`.
    ;;
    ;; Return a list of tokens.
  (let ([text (source-text source)])
    (let loop ([start 0])
      (let-values ([(token end) (parse-token source text start #f)])
        (if (eqv? token 'eof)
            '()
          (cons token (loop end)))))))

;;; String parsing routines.
(define (parse-string source text start spaced?)
  ;; Parse a "-delimited string, beginning at `start`. Build a token of
  ;; 'string kind, with the string contents as value, and the specified
  ;; `spaced?` property.
  ;;
  ;; Return two values: the token, and the next position.
  (let-values ([(content next) (parse-string-content text start #\")])
    (values (make-token (make-location source start) 'string content spaced?) next)))

(define (parse-symbol source text start spaced?)
  ;; Parse a `-delimited symbol, beginning at `start`. Build a token of
  ;; 'symbol kind, with the string contents as value, and the specified
  ;; `spaced?` property.
  ;;
  ;; Return two values: the token, and the next position.
  (let-values ([(content next) (parse-string-content text start #\`)])
    (values (make-token (make-location source start) 'symbol content spaced?) next)))

(define (parse-string-content text start close)
  ;; Parse the contents of a string or symbol beginning at `start`, delimited
  ;; by the `close` character. Interpret backslash escapes.
  ;;
  ;; Return two values: the string contents, and the next position after the
  ;; `close` character.
  (let ([len (string-length text)])
    (let loop ([i (+ 1 start)] [chars '()])
      (if (>= i len)
          (raise-eof-error 'parse-string "Found EOF in the middle of string")
        (let ([char (string-ref text i)])
          (cond [(eqv? char close)
                 (values (list->string (reverse chars)) (+ 1 i))]
                [(eqv? char #\\)
                 (let ([next (string-ref text (+ i 1))])
                   ;; TODO: escapes
                   (loop (+ i 2) (cons next chars)))]
                [else (loop (+ i 1) (cons char chars))]))))))

(define (next-token tokens skipnl?)
  ;; Find the next token in a list of token. Skip newlines when `skipnl?` is true.
  ;;
  ;; Return two values: the next token (or #f if no more tokens available),
  ;; and the rest of the list.
  (cond [(null? tokens) (values #f tokens)]
        [(and skipnl? (eqv? (token-value (car tokens)) #\newline))
         (next-token (cdr tokens) skipnl?)]
        [else (values (car tokens) (cdr tokens))]))

(define (end-of-phrase? token skipnl?)
  ;; True if a given token delimits a phrase.
  (or (not token)
      (member (token-value token) '(#\; #\, #\] #\) #\}))
      (and (eqv? (token-value token) #\newline) (not skipnl?))))
