#!/bin/bash

read_block() {
    local _var="$1"
    local _text=""
    while IFS= read -r line; do
        _text+="$line"$'\n'
    done
    printf -v "$_var" "$_text"
}

TMP_CODE="/tmp/test.fen"
TMP_OUTPUT="/tmp/test.out"
TMP_EXPECTED="/tmp/test.expected"

code() { ((tests++)); read_block code; }
output() { read_block expected_output; }
test() { test_name="$1"; }

tests=0
success=0
failed=0

run() {
    echo "Running test '$test_name'..."
    echo -n "$code" >"$TMP_CODE"
    echo -n "$expected_output" >"$TMP_EXPECTED"
    ./fenius.scm "$TMP_CODE" >"$TMP_OUTPUT"
    if diff "$TMP_EXPECTED" "$TMP_OUTPUT"; then
        (( success++ ))
    else
        (( failed++ ))
    fi
}

#####################

cd "$(dirname "$0")/../src"

test "fac"

code <<"EOF"
let fac(n) = {
    if n <= 0 {
        1
    } else {
        n * fac(n-1)
    }

}

print(fac(5))
print(fac(n=5))
{
    let arguments = [5]
    print(fac(*arguments))
}
EOF

output <<"EOF"
120
120
120
EOF

run

######################

test "fib"

code <<"EOF"
let fib(n) = {
    if n <= 1 {
        1
    } else {
       fib(n-1) + fib(n-2)
    }
}

print(fib(10))

let fib_aux(n, lesser, greater) = {
    if n == 0 {
        greater
    } else {
        fib_aux(n-1, greater, lesser+greater)
    }
}

print(fib_aux(40, 0, 1))
EOF

output <<"EOF"
89
165580141
EOF

run

######################

test "macros"
code <<"EOF"
let just = Macro(fn (expr) expr.constituents[1])

print(just 2+3)


let foreach(items, f) = {
    if items != [] {
        f(items.first)
        foreach(items.rest, f)
    }
}


# hel> quote(foreach(items, fn (var) body))
# Call(Identifier(`foreach`), [Identifier(`items`), Phrase([Identifier(`fn`), Identifier(`var`), Identifier(`body`)])])
# hel> quote(for var in items body)
# Phrase([Identifier(`for`), Identifier(`var`), Identifier(`in`), Identifier(`items`), Identifier(`body`)])

let for = Macro(fn (source) {
    let var = source.constituents[1]
    let items = source.constituents[3]
    let body = source.constituents[4]

    Call(false, Identifier(false, `foreach`),
                [items, Phrase(false, [Identifier(false, `fn`), var, body])])
})

for x in [1, 2, 3] {
    print(x)
}
EOF

output <<EOF
5
1
2
3
EOF

run

###########

test "macros+match"
code <<"EOF"
let foreach(items, f) = {
    if items != [] {
        f(items.first)
        foreach(items.rest, f)
    }
}

let for = Macro(fn (ast) {
    match ast {
        ast_match(for _(var) in _(items) _(body)) => {
            ast_gen(foreach(_(items), fn (_(var)) _(body)))
        }
    }
})

for x in [1, 2, 3] {
    print(x)
}
EOF

output <<EOF
1
2
3
EOF

run

######################

test "records"

code <<EOF
record Person(name, age)

method Person.say(thing) = {
    print(this.name ++ " says '" ++ thing ++ "'")
}

let p = Person("Hildur", 23)
p.say("Hello!")

record Person > Worker(job)

let w = Worker("Olafur", 24, "programmer")
w.say("Hello!")
EOF

output <<EOF
Hildur says 'Hello!'
Olafur says 'Hello!'
EOF

run

#######################

test "loops"
# XXX: This test is a bit pointless. Should use with foreach to
# demonstrate actual non-local exit.

code <<EOF
let product(items) = {
    do block(result=1, items=items) {
        if items == [] {
            return result from block
        } elif items.first == 0 {
            return 0 from block
        } else {
            print("Looking at ", items.first)
            redo block(result = result*items.first, items = items.rest)
        }
    }
}

print("product 1 = ", product([1,2,3,4,5]))
print("product 2 = ", product([1,2,0,4,5]))
EOF

output <<EOF
Looking at 1
Looking at 2
Looking at 3
Looking at 4
Looking at 5
product 1 = 120
Looking at 1
Looking at 2
product 2 = 0
EOF

run

######################

test 'exceptions'

code <<EOF

record Error(message)

let test_exceptions(x) = {
    try {
        raise x
    } except err : Error {
        print("Got an error: " ++ err.message)
    } except err : Object {
        print("Got something else")
    } finally {
        print("Fortunately it's over")
    }
}

test_exceptions(Error("no way"))
test_exceptions(42)
EOF

output <<EOF
Got an error: no way
Fortunately it's over
Got something else
Fortunately it's over
EOF

run

######################

test 'trailing newlines'

code <<EOF
print("Hello, world")

EOF

output <<EOF
Hello, world
EOF

run

######################

# TODO: test dicts, make order predictable somehow.



if (( success + failed != tests)); then
    echo "ERROR: forgot 'run' in some test!" >&2
fi

echo "$success tests passed, $failed failed."
(( failed )) && exit 1
